import os

import googlemaps

from utils import timex
from utils.cache import cache

CACHE_NAME, CACHE_TIMEOUT = 'GoogleMaps', timex.SECONDS_IN.YEAR


def get_api_key():
    return os.environ['GOOGLE_API_KEY']


class GoogleMaps:
    def __init__(self):
        gmaps_api_key = get_api_key()
        self.api = googlemaps.Client(key=gmaps_api_key)

    def __get_geocode__(self, address):
        @cache(CACHE_NAME, CACHE_TIMEOUT)
        def get_geocode_inner(address):
            return self.api.geocode(address)

        return get_geocode_inner(address)

    def __get_reverse_geocode__(self, latlng):
        @cache(CACHE_NAME, CACHE_TIMEOUT)
        def get_reverse_geocode_inner(latlng):
            return self.api.reverse_geocode(latlng)

        return get_reverse_geocode_inner(latlng)

    def get_latlng(self, address):
        geocode = self.__get_geocode__(address)
        location = geocode[0]['geometry']['location']
        return [location['lat'], location['lng']]

    def get_address(self, latlng):
        rev_geocode = self.__get_reverse_geocode__(latlng)
        return rev_geocode[0]['formatted_address']

    def get_address_details(self, latlng):
        rev_geocode = self.__get_reverse_geocode__(latlng)
        if not rev_geocode:
            return None
        first_rev_geocode_item = rev_geocode[0]

        street_number = None
        street = None
        sub_city = None
        city = None
        postal_code = None
        district = None
        country = None
        plus_code = None

        for ac in first_rev_geocode_item['address_components']:
            ac_types = ac['types']
            if 'street_number' in ac_types:
                street_number = ac['long_name']
            if 'route' in ac_types:
                street = ac['long_name']
            if 'sublocality' in ac_types:
                sub_city = ac['long_name']
            if 'locality' in ac_types:
                city = ac['long_name']
            if 'postal_code' in ac_types:
                postal_code = ac['long_name']

            if 'administrative_area_level_2' in ac_types:
                district = ac['long_name']
            if 'administrative_area_level_1' in ac_types:
                province = ac['long_name']
            if 'country' in ac_types:
                country = ac['long_name']

            if 'plus_code' in ac_types:
                plus_code = ac['long_name']

        details = dict(
            street_number=street_number,
            plus_code=plus_code,
            street=street,
            sub_city=sub_city,
            city=city,
            postal_code=postal_code,
            district=district,
            province=province,
            country=country,
            formatted_address=first_rev_geocode_item['formatted_address'],
        )
        return details



import time


class TIMEZONE_OFFSET:
    LK = -19_800
    GMT = 0


class SECONDS_IN:
    MINUTE = 60
    HOUR = 3600
    DAY = 86400
    WEEK = 7 * 86400
    FORTNIGHT = 14 * 86400
    MONTH = 365.25 * 4 * 86400 / 12
    QTR = 365.25 * 86400 / 4
    YEAR = 365.25 * 86400


class AVG_DAYS_IN:
    WEEK = 7
    MONTH = 365.25 / 12
    YEAR = 365.25


class TimeDelta:
    def __init__(self, dut):
        self.dut = dut

    def __eq__(self, other) -> bool:
        return self.dut == other.dut

    def humanize(self):
        dut = self.dut
        if dut < 60:
            return f'{dut:.0f}sec'
        dut /= 60
        if dut < 60:
            return f'{dut:.0f}min'
        dut /= 60
        if dut < 24:
            return f'{dut:.0f}hr'
        dut /= 24
        return f'{dut:.0f}day'


class Time:
    def __init__(self, ut):
        self.ut = ut

    def __eq__(self, other) -> bool:
        return self.ut == other.ut

    def __sub__(self, other) -> TimeDelta:
        return TimeDelta(self.ut - other.ut)

    @staticmethod
    def now():
        return Time(time.time())


class TimeFormat:
    def __init__(self, format_str: str, timezone_offset=0):
        self.format_str = format_str
        self.timezone_offset = timezone_offset

    @property
    def dut_timezone(self):
        return time.timezone - self.timezone_offset

    def parse(self, time_str: str) -> Time:
        ut = (
            time.mktime(time.strptime(time_str, self.format_str))
            - self.dut_timezone
        )
        return Time(ut)

    def stringify(self, t: Time) -> str:
        return time.strftime(
            self.format_str, time.localtime(t.ut + self.dut_timezone)
        )

    


def latlng_to_xy(latlng):
    lat, lng = latlng
    return [lng, lat]


def xy_to_latlng(xy):
    x, y = xy
    return [y, x]


def get_bbox(xy_list):
    max_x, max_y = min_x, min_y = xy_list[0]
    for [x, y] in xy_list[1:]:
        min_x = min(min_x, x)
        max_x = max(max_x, x)
        min_y = min(min_y, y)
        max_y = max(max_y, y)

    x_span = max_x - min_x
    y_span = max_y - min_y

    return [
        [min_x, min_y],
        [max_x, max_y],
        [x_span, y_span],
    ]


def get_func_transform(width, height, padding, xy_list):
    [
        [min_x, min_y],
        [max_x, max_y],
        [x_span, y_span],
    ] = get_bbox(xy_list)

    padded_width, padded_height = width - padding * 2, height - padding * 2
    r = (x_span / padded_width) / (y_span / padded_height)
    if r > 1:
        padded_height /= r
    else:
        padded_width *= r

    padding_width = (width - padded_width) / 2
    padding_height = (height - padded_height) / 2

    def func_transform(xy):
        [x, y] = xy
        px = (x - min_x) / x_span
        py = (y - min_y) / y_span

        x = (px) * padded_width + padding_width
        y = (1 - py) * padded_height + padding_height
        return [x, y]

    return func_transform



"""JSON utils."""
import base64
import json

from utils import filex


def read(file_name):
    """Read JSON from file.

    Args:
        file_name (str): file name

    Returns:
        Parsed JSON data

    """
    return json.loads(filex.read(file_name))


def write(file_name, data):
    """Write data as JSON to file.

    Args:
        file_name (str): file name
        data: data as serializable object

    """
    filex.write(file_name, json.dumps(data, indent=2))


def serialize(data):
    if isinstance(data, bytes):
        return {
            'type': 'bytes',
            'data': base64.b64encode(data).decode('ascii'),
        }

    return {
        'type': None,
        'data': data,
    }


def deserialize(data):
    data_type = data['type']
    data_data = data['data']

    if data_type == 'bytes':
        return base64.b64decode(data_data)

    return data_data



import csv
import json
import os
import zipfile

DIALECT = 'excel'
DELIMITER_CSV = ','
DELIMITER_TSV = '\t'
DELIM_LINE = '\n'


class File:
    def __init__(self, file_name):
        self.file_name = file_name

    def read(self):
        with open(self.file_name, 'r') as fin:
            content = fin.read()
            fin.close()
        return content

    def readBinary(self):
        with open(self.file_name, 'rb') as fin:
            content = fin.read()
            fin.close()
        return content

    def write(self, content):
        with open(self.file_name, 'w') as fout:
            fout.write(content)
            fout.close()

    def read_lines(self):
        content = File.read(self)
        return content.split(DELIM_LINE)

    def write_lines(self, lines):
        content = DELIM_LINE.join(lines)
        File.write(self, content)


class JSONFile(File):
    def read(self):
        content = File.read(self)
        return json.loads(content)

    def write(self, data):
        content = json.dumps(data, indent=2)
        File.write(self, content)


class XSVFile(File):
    def __init__(self, file_name, delimiter):
        File.__init__(self, file_name)
        self.delimiter = delimiter

    def read(self):
        csv_lines = File.read_lines(self)

        data_list = []
        field_names = None
        reader = csv.reader(
            csv_lines,
            dialect=DIALECT,
            delimiter=self.delimiter,
        )
        for row in reader:
            if not field_names:
                field_names = row
            else:
                data = dict(
                    zip(
                        field_names,
                        row,
                    )
                )
                if data:
                    data_list.append(data)
        return data_list

    def write(self, data_list):
        with open(self.file_name, 'w') as fout:
            writer = csv.writer(
                fout,
                dialect=DIALECT,
                delimiter=self.delimiter,
            )

            field_names = list(data_list[0].keys())
            writer.writerow(field_names)
            writer.writerows(
                list(
                    map(
                        lambda data: list(
                            map(
                                lambda field_name: data[field_name],
                                field_names,
                            )
                        ),
                        data_list,
                    )
                ),
            )
            fout.close()


class CSVFile(XSVFile):
    def __init__(self, file_name):
        return XSVFile.__init__(self, file_name, DELIMITER_CSV)


class TSVFile(XSVFile):
    def __init__(self, file_name):
        return XSVFile.__init__(self, file_name, DELIMITER_TSV)


class Zip:
    def __init__(self, file_name):
        self.file_name = file_name

    @property
    def zip_file_name(self):
        return self.file_name + '.zip'

    @property
    def arc_name(self):
        return os.path.basename(os.path.normpath(self.file_name))

    @property
    def dir_zip(self):
        return os.path.dirname(os.path.normpath(self.file_name))

    def zip(self, skip_delete=False):
        assert os.path.exists(self.file_name)
        with zipfile.ZipFile(
            self.zip_file_name,
            mode='w',
            compression=zipfile.ZIP_DEFLATED,
        ) as zip_file:
            zip_file.write(self.file_name, arcname=self.arc_name)
            assert os.path.exists(self.zip_file_name)

        if not skip_delete:
            os.remove(self.file_name)
            assert not os.path.exists(self.file_name)

    def unzip(self, skip_delete=False):
        assert os.path.exists(self.zip_file_name)
        with zipfile.ZipFile(self.zip_file_name) as zip_file:
            zip_file.extractall(self.dir_zip)
            assert os.path.exists(self.file_name)

        if not skip_delete:
            os.remove(self.zip_file_name)
            assert not os.path.exists(self.zip_file_name)



from utils.File import CSVFile, File, JSONFile, TSVFile, XSVFile, Zip
from utils.Git import Git
from utils.GoogleMaps import GoogleMaps
from utils.Time import (
    AVG_DAYS_IN,
    SECONDS_IN,
    TIMEZONE_OFFSET,
    Time,
    TimeDelta,
    TimeFormat,
)
from utils.www import WWW
